import { Request, Response } from 'express';
import { v4 as uuidv4 } from 'uuid';
import { decamelizeKeys, camelizeKeys } from 'xcase';
import { connection } from '../../config/ormconfig';
import { criticalPrefix, sessionUsedByOtherUser } from '../../json/en.json';
import { log } from '../../utils/LogError';
import { LogErrorCodes } from '../../utils/ErrorCodes';
import { logError } from '../../utils/ErrorLogger';
import { ResponseStatus } from '../../utils/ResponseStatus';
import playerControllerHelpers from './helpers/playerControllerHelpers';
import asyncSaveSubmitHelper from './helpers/asyncSaveSubmitHelper';
import { getRedisInstance } from '../../redisManager';
import { promisify } from 'util';
import { TimeLogger } from '../../utils/timeLogger';

const {
    studentPlayerViewSchema,
    saveSubmitInputSchema,
    checkStudentExists,
    storeStudentData,
    getAssignmentDataByReferenceId,
    doesStudentHasAccessOrIsAuthorized,
    fetchAssignmentQuestionMap,
    storeAssignmentQuestionMaps,
    extractQuestionPartsMap,
    getActivityData,
} = playerControllerHelpers;

const {
    saveSubmitResponse,
    checkLiveClassStatus,
    updateQueueRequestCount,
    pushMessageInQueue,
} = asyncSaveSubmitHelper;

class PlayerController {
    public async studentPlayerView(req: Request, res: Response) {
        log('studentPlayerView', req.url, 'grading view api');
        const timeLogger = new TimeLogger('studentPlayerView-loadtest', [uuidv4()]);
        const { body, headers: { user_id: userId } } = req;
        const { activityId, activityTemplateId, organisationId, retrieve_tags = false,
            studentId = '', sessionId = '' } = body;
        log('studentPlayerView', 'studentPlayerViewSchema', 'validate request body data');
        const data = { activityId, activityTemplateId,
            retrieve_tags,
            studentId, sessionId, organisationId };
        const { error: validationError } = studentPlayerViewSchema(decamelizeKeys(data));
        if (validationError) {
            console.error(validationError);
            logError(
                LogErrorCodes.INVALID_REQUEST,
                `${criticalPrefix} ${validationError.details[0].message}`,
                'studentPlayerView',
            );
            return res.status(ResponseStatus.INVALID_REQUEST).json({
                message: validationError.details[0].message,
                success: false,
            });
        }
        connection
            .then(async (conn) => {
                let activityData;
                log(
                    'studentPlayerView',
                    'fetchAssessmentCorrectAnswerView',
                    'fetch activity data (inter service call)',
                );
                const assessmentParamas = { activityTemplateId, userId, organisationId, timeLogger };
                activityData = await getActivityData(assessmentParamas);
                log(
                    'studentPlayerView',
                    'checkStudentExists',
                    'check if student assignment mapping data exists',
                );
                const responseParams = { conn, ...data, assessmentId: activityData.id };
                let isStudentExist = await checkStudentExists(responseParams);
                if (!isStudentExist) {
                    log('studentPlayerView', 'getAssignmentDataByReferenceId', '');
                    const selectFields = ['assignment.id', 'assignment.assignmentId'];
                    const assignmentParams = { conn, organisationId, selectFields, assignmentReferenceId: activityId };
                    const assignment = await getAssignmentDataByReferenceId(assignmentParams);
                    const studentParams = { studentId, sessionId, assignmentId: assignment.id, transaction: conn };
                    isStudentExist = await storeStudentData(studentParams);
                }
                log(
                    'studentPlayerView',
                    'fetchAssignmentQuestionMap',
                    'fetch assignment question mapping data',
                );
                const assignmentQuestionMaps = await fetchAssignmentQuestionMap(responseParams) || [];
                const questionQuestionPartMaps = extractQuestionPartsMap(assignmentQuestionMaps);
                const items = activityData?.items;
                const storeParams = {
                    conn, sessionId, items, assignmentQuestionMaps,
                    questionQuestionPartMaps, studentAssignmentId: isStudentExist.id,
                };
                log(
                    'studentPlayerView', 'storeAssignmentQuestionMaps', 'store assignment question mapping',
                );
                const newQuestionParts = await storeAssignmentQuestionMaps(storeParams);

                const questionPartsResponse = newQuestionParts.length ? newQuestionParts : questionQuestionPartMaps;
                if (activityData) {
                    activityData.questions = activityData?.items;
                    delete activityData?.items;
                    activityData.existingSession = isStudentExist ? true : false;
                    for (const question of activityData?.questions) {
                        question.questionParts = question?.questions;
                        delete question?.questions;
                        if (!retrieve_tags) {
                            question.questionTags = [];
                        }
                        for (const questionPart of question?.questionParts) {
                            const response = questionPartsResponse.find(
                                r => r.questionPartId === questionPart.id && r.questionId === question.questionId,
                            );
                            questionPart.response = {
                                responseId: response?.responseId,
                                questionPartId: questionPart.id,
                                response: response?.response || [],
                            };
                            questionPart.isAttempted = response?.isAttempted || false;
                        }
                    }
                }

                res.status(ResponseStatus.STATUS_OK).json(
                    camelizeKeys({
                        playerData: activityData,
                        success: true,
                    }),
                );
            })
            .catch((error) => {
                console.error(error);
                logError(
                    LogErrorCodes.RESPONSE_ERROR,
                    `${criticalPrefix} ${error.message}`,
                    'studentPlayerView',
                );
                return res.status(ResponseStatus.INTERNAL_SERVER_ERROR).json({
                    success: false,
                    message: error.message,
                });
            });
    }

    public async asyncSaveOrSubmitAssignmentResponse(req: Request, res: Response) {
        log('asyncSaveOrSubmitAssignmentResponse', req.url, 'save or submit student response');
        const requestId = uuidv4();
        const { body, params } = req;
        const { submitOrSave } = params;
        const { activityId, activityTemplateId, questions, organisationId, sessionId = '',
            studentId = '', windowNonce = '', isCoPresenter = false } = body;
        log(
            'asyncSaveOrSubmitAssignmentResponse',
            'saveSubmitInputSchema',
            `validate request body data with requestId-${requestId}`,
        );
        const data = { activityId, activityTemplateId, sessionId, studentId, questions, organisationId };
        const { error: validationError } = saveSubmitInputSchema(decamelizeKeys(data));
        if (validationError) {
            console.error(validationError);
            logError(
                LogErrorCodes.INVALID_REQUEST,
                `${criticalPrefix} ${validationError.details[0].message}`,
                'asyncSaveOrSubmitAssignmentResponse',
            );
            return res.status(ResponseStatus.INVALID_REQUEST).json({
                message: validationError.details[0].message,
                success: false,
            });
        }

        // To limit only one co-presenter per student in an item.
        if (isCoPresenter) {
            const redisClient = getRedisInstance();
            const asyncSet = promisify(redisClient.set).bind(redisClient);
            const asyncGet = promisify(redisClient.get).bind(redisClient);
            const redisKey = `co-presenter:${activityTemplateId}:${studentId}`;

            const [currentNonce, currentInitTime] = (await asyncGet(redisKey))?.split(':') || [];
            const [nonce, initTime] = windowNonce.split(':');

            // If co-presenter exists and window is different and new window is older then current window;
            // Then return session being used error.
            if (currentNonce && currentInitTime && currentNonce !== nonce && +initTime < +currentInitTime) {
                console.error(sessionUsedByOtherUser);
                logError(
                    LogErrorCodes.INVALID_REQUEST,
                    `${criticalPrefix} ${sessionUsedByOtherUser}`,
                    'asyncSaveOrSubmitAssignmentResponse',
                );
                return res.status(ResponseStatus.CONFLICT).json({
                    message: sessionUsedByOtherUser,
                    errorCode: 1540,
                    success: false,
                });
            }

            // Else set window data in redis and proceed further.
            asyncSet(redisKey, windowNonce, 'EX', 6);
        }

        connection
            .then(async (conn) => {
                const isSubmitOrSave = parseInt(submitOrSave, 10);
                const selectFields = ['assignment.id', 'assignment.assignmentId'];
                const assignmentParams = { conn, organisationId, selectFields, assignmentReferenceId: activityId };

                log('asyncSaveOrSubmitAssignmentResponse', 'getAssignmentDataByReferenceId', `requestId-${requestId}`);
                const assignment = await getAssignmentDataByReferenceId(assignmentParams);
                const { assignmentId } = assignment;
                const studentAuthorizedParams = { studentId, assignmentId };

                log('asyncSaveOrSubmitAssignmentResponse', 'doesStudentHasAccessOrIsAuthorized', `requestId-${requestId}`);
                const studentHasAccess = await doesStudentHasAccessOrIsAuthorized(studentAuthorizedParams);
                const { studentAssignmentId } = studentHasAccess;
                if (process.env.PLAYER__ENABLE_LAMBDA_FOR_SAVE_SUBMIT === 'true') {
                    // Update with live class redis check
                    log('checkLiveClassStatus', '', `called from async save/submit with requestId-${requestId}`);
                    const isLiveClassEnabled = await checkLiveClassStatus(activityId, organisationId);

                    log('updateQueueRequestCount', '', `called from async save/submit with requestId-${requestId}`);
                    const isSimpleQueueRequest = await updateQueueRequestCount(
                        studentAssignmentId, assignmentId, isLiveClassEnabled,
                    );

                    const message = {
                        studentHasAccess, questions, isSubmitOrSave, activityId, studentId,
                        organisationId, isSimpleQueueRequest, requestId, sessionId,
                    };

                    log('pushMessageInQueue', '', `called from async save/submit with requestId-${requestId}`);
                    await pushMessageInQueue(isSimpleQueueRequest, message, studentAssignmentId, assignmentId);

                    return res.status(200).send({ success: true });
                }

                const saveAssignmentResponseDataParams = {
                    studentHasAccess, questions, isSubmitOrSave, activityId, studentId,
                    organisationId, requestId, sessionId,
                };
                log('asyncSaveOrSubmitAssignmentResponse', 'saveSubmitResponse', `requestId-${requestId}`);
                const { success, message } = await saveSubmitResponse(saveAssignmentResponseDataParams, requestId);
                if (!success) {
                    console.error(message);
                    logError(
                        LogErrorCodes.RESPONSE_ERROR,
                        `${criticalPrefix} ${message}`,
                        'asyncSaveOrSubmitAssignmentResponse',
                    );
                    return res.status(ResponseStatus.INTERNAL_SERVER_ERROR).json({
                        message,
                        success: false,
                    });
                }
                res.status(ResponseStatus.STATUS_OK).json({
                    success: true,
                });
            })
            .catch((error) => {
                console.error(error);
                logError(
                    LogErrorCodes.RESPONSE_ERROR,
                    `${criticalPrefix} ${error.message}`,
                    'asyncSaveOrSubmitAssignmentResponse',
                );
                return res.status(ResponseStatus.INTERNAL_SERVER_ERROR).json({
                    success: false,
                    message: error.message,
                });
            });
    }
}

export default PlayerController;
