import {
    mockedQueryBuilder,
    mockedRepository,
} from '../__mocks__/typeorm.mock';
import supertest from 'supertest';
import { mocked } from 'ts-jest/utils';
import axios from 'axios';
import urlJoin from 'url-join';
import redisMock from 'redis-mock';
import app from '../../../app';
import { ResponseStatus } from '../../../utils/ResponseStatus';
import mockData from '../__mocks__/mockData';

const request = supertest(app);
jest.mock('axios');
const mockAxios = mocked(axios, true);
const redisClient = redisMock.createClient();
jest.mock('../../../redisManager', () => ({
    getRedisInstance: jest.fn(() => redisClient)}),
);

const {
    activityMockedData,
    saveSubmitInputData,
    assignmentStudentMapMockData,
} = mockData;
const prefix = process.env.ROUTE_PREFIX || '';

describe('PlayerController', () => {
    let response;
    beforeAll(async () => {
        mockedQueryBuilder.getRawOne.mockClear();
        mockedRepository.save.mockClear();
        mockedQueryBuilder.getOne.mockClear();
        mockedQueryBuilder.getRawMany.mockClear();
        mockedQueryBuilder.execute.mockClear();
        mockedQueryBuilder.getMany.mockClear();
    });

    afterAll(async () => {
        jest.clearAllMocks();
        mockedQueryBuilder.getRawOne.mockClear();
        mockedRepository.save.mockClear();
        mockedQueryBuilder.getOne.mockClear();
        mockedQueryBuilder.getRawMany.mockClear();
        mockedQueryBuilder.execute.mockClear();
        mockedQueryBuilder.getMany.mockClear();
    });

    const organisationId = '44a4f44a-44a4-4444-a4a4-4aaa4aa44aa4';
    describe('studentPlayerView', () => {
        const activityDataPath = urlJoin(process.env.ACTIVITY_SERVICE, '/activity/view/correctAnswer');
        const requestPath = `${prefix}/player/student`;
        let mockPost;
        describe('when valid input are passed and student already exists', () => {
            const activityData = activityMockedData();
            const input = {
                organisationId,
                activityId: 'f86cfbcd-254e-4f3c-add7-d2ab2fffc2e0',
                activityTemplateId: '263d353-ca6c-20-60-1a2c60e21ec3',
                studentId: 'c96eadb2-8e0b-49e7-ada2-0645734789a7',
                sessionId: '80c8312d-af7d-48de-8771-5b33ff27632d',
            };
            beforeAll(async () => {
                redisClient.flushall('ASYNC');
                mockPost = mockAxios.post.mockImplementation((urlpath) => {
                    if (urlpath.includes('/activity/view/correctAnswer')) {
                        return Promise.resolve({ data: { activityData } });
                    }
                    return Promise.resolve({ data: { success: true } });
                });
                mockedQueryBuilder.getRawOne.mockResolvedValueOnce(assignmentStudentMapMockData[0]);
                mockedQueryBuilder.getMany.mockResolvedValueOnce([{
                    qustion_part_id: 10,
                    response_id: 11,
                    questionParts: [],
                }]);
                mockedRepository.save.mockResolvedValueOnce({});
                mockedQueryBuilder.execute.mockResolvedValueOnce({});
                response = await request.post(requestPath).send({ ...input });
            });
            afterAll(async () => {
                mockAxios.post.mockClear();
                jest.clearAllMocks();
                mockedQueryBuilder.getRawOne.mockReset();
                mockedQueryBuilder.getRawOne.mockClear();
                mockedQueryBuilder.getMany.mockReset();
                mockedQueryBuilder.getMany.mockClear();
                mockedRepository.save.mockReset();
                mockedRepository.save.mockClear();
                mockedQueryBuilder.execute.mockReset();
                mockedQueryBuilder.execute.mockClear();
            });
            it('should call remote post apis with correct parameters', () => {
                const firstPostRequest = mockPost.mock.calls[0];
                expect(firstPostRequest[0]).toEqual(activityDataPath);
                expect(firstPostRequest[1]).toMatchObject({
                    activityReference: input.activityTemplateId,
                });
                expect(firstPostRequest[2]).toEqual(expect.any(Object));
            });
            it('should return correct response', () => {
                expect(response.status).toBe(200);
                const { playerData } = response.body;
                const properties = [
                    'createdAt', 'createdBy', 'description', 'displayId', 'id', 'questions', 'referenceId',
                    'status', 'statusId', 'tags', 'title', 'totalPoints', 'type', 'updatedAt', 'version',
                ];
                expect(response.body).toHaveProperty('playerData');
                expect(response.body).toHaveProperty('success');
                for (const property of properties) {
                    expect(playerData).toHaveProperty(property);
                }
            });
        });
        describe('when valid input are passed and no questionMap', () => {
            const activityData = activityMockedData();
            const input = {
                organisationId,
                activityId: 'f86cfbcd-254e-4f3c-add7-d2ab2fffc2e0',
                activityTemplateId: '263d353-ca6c-20-60-1a2c60e21ec3',
                studentId: 'c96eadb2-8e0b-49e7-ada2-0645734789a7',
                sessionId: '80c8312d-af7d-48de-8771-5b33ff27632d',
            };
            beforeAll(async () => {
                redisClient.flushall('ASYNC');
                mockPost = mockAxios.post.mockImplementation((urlpath) => {
                    if (urlpath.includes('/activity/view/correctAnswer')) {
                        return Promise.resolve({
                            data: { activityData: { ...activityData, items: [] } },
                        });
                    }
                    return Promise.resolve({ data: { success: true } });
                });
                mockedQueryBuilder.getRawOne.mockResolvedValueOnce(assignmentStudentMapMockData[0]);
                mockedQueryBuilder.getMany.mockResolvedValueOnce([{
                    qustion_part_id: 10,
                    response_id: 11,
                    questionParts: [],
                }]);
                mockedRepository.save.mockResolvedValueOnce({});
                mockedQueryBuilder.execute.mockResolvedValueOnce({});
                response = await request.post(requestPath).send({ ...input });
            });
            afterAll(async () => {
                mockAxios.post.mockClear();
                jest.clearAllMocks();
                mockedQueryBuilder.getRawOne.mockReset();
                mockedQueryBuilder.getRawOne.mockClear();
                mockedQueryBuilder.getMany.mockReset();
                mockedQueryBuilder.getMany.mockClear();
                mockedRepository.save.mockReset();
                mockedRepository.save.mockClear();
                mockedQueryBuilder.execute.mockReset();
                mockedQueryBuilder.execute.mockClear();
            });
            it('should call remote post apis with correct parameters', () => {
                const firstPostRequest = mockPost.mock.calls[0];
                expect(firstPostRequest[0]).toEqual(activityDataPath);
                expect(firstPostRequest[1]).toMatchObject({
                    activityReference: input.activityTemplateId,
                });
                expect(firstPostRequest[2]).toEqual(expect.any(Object));
            });
            it('should return correct response', () => {
                expect(response.status).toBe(200);
                const { playerData } = response.body;
                const properties = [
                    'createdAt', 'createdBy', 'description', 'displayId', 'id', 'questions', 'referenceId',
                    'status', 'statusId', 'tags', 'title', 'totalPoints', 'type', 'updatedAt', 'version',
                ];
                expect(response.body).toHaveProperty('playerData');
                expect(response.body).toHaveProperty('success');
                for (const property of properties) {
                    expect(playerData).toHaveProperty(property);
                }
            });
        });
        describe('when valid input are passed and student not exists', () => {
            const activityData = activityMockedData();
            const inputBody = {
                organisationId,
                activityId: 'f86cfbcd-254e-4f3c-add7-d2ab2fffc2e0',
                activityTemplateId: '263d353-ca6c-20-60-1a2c60e21ec3',
                studentId: '7ece621b-23a8-4006-8c52-030fe029c666',
                sessionId: '805bf693-1d2d-4cae-adfa-54ec1518ebea',
            };
            beforeAll(async () => {
                mockAxios.post.mockImplementation((urlpath) => {
                    if (urlpath.includes('/activity/view/correctAnswer')) {
                        return Promise.resolve({ data: { activityData } });
                    }
                    return Promise.resolve({ data: { success: true } });
                });
                mockedQueryBuilder.getRawOne.mockResolvedValueOnce(false);
                mockedQueryBuilder.getOne.mockResolvedValueOnce({
                    id: 1,
                    assignmentId: 1,
                });
                mockedRepository.save.mockResolvedValueOnce(assignmentStudentMapMockData[0])
                .mockResolvedValueOnce({});
                mockedQueryBuilder.getMany.mockResolvedValueOnce([{
                    qustion_part_id: 10,
                    response_id: 11,
                    questionParts: [],
                }]);
                mockedQueryBuilder.execute.mockResolvedValueOnce({});
                response = await request.post(requestPath).send(inputBody);
            });
            afterAll(() => {
                mockAxios.post.mockClear();
                jest.clearAllMocks();
                mockedQueryBuilder.getRawOne.mockReset();
                mockedQueryBuilder.getRawOne.mockClear();
                mockedQueryBuilder.getOne.mockReset();
                mockedQueryBuilder.getOne.mockClear();
                mockedQueryBuilder.getMany.mockReset();
                mockedQueryBuilder.getMany.mockClear();
                mockedRepository.save.mockReset();
                mockedRepository.save.mockClear();
                mockedQueryBuilder.execute.mockReset();
                mockedQueryBuilder.execute.mockClear();
            });
            it('should return correct response', () => {
                expect(response.status).toBe(200);
                const { playerData } = response.body;
                const properties = [
                    'createdAt', 'createdBy', 'description', 'displayId', 'id', 'questions', 'referenceId',
                    'status', 'statusId', 'tags', 'title', 'totalPoints', 'type', 'updatedAt', 'version',
                ];
                expect(response.body).toHaveProperty('playerData');
                expect(response.body).toHaveProperty('success');
                for (const property of properties) {
                    expect(playerData).toHaveProperty(property);
                }
            });
        });
        describe('when invalid input are passed', () => {
            const input = {
                organisationId,
                activityId: 'f86cfbcd-254e-4f3c-add7-d2ab2fffc2e0',
                activityTemplateId: '263d353-ca6c-20-60-1a2c60e21ec3',
            };
            beforeAll(async () => {
                response = await request.post(requestPath).send({ ...input });
            });
            it('should return correct response', () => {
                expect(response.status).toBe(422);
                expect(response.body).toHaveProperty('message', '"student_id" is not allowed to be empty');
                expect(response.body).toHaveProperty('success', false);
            });
        });
        describe('when error', () => {
            const inputData = {
                organisationId,
                activityId: 'f86cfbcd-254e-4f3c-add7-d2ab2fffc2e0',
                activityTemplateId: '263d353-ca6c-20-60-1a2c60e21ec3',
                studentId: 'c96eadb2-8e0b-49e7-ada2-0645734789a7',
                sessionId: '80c8312d-af7d-48de-8771-5b33ff27632d',
            };
            beforeAll(async () => {
                mockAxios.post.mockImplementation((urlpath) => {
                    if (urlpath.includes('/activity/view/correctAnswer')) {
                        throw new Error();
                    }
                    return Promise.resolve({ data: { success: true } });
                });
                response = await request.post(requestPath).send(inputData);
            });
            afterAll(async () => {
                mockAxios.post.mockClear();
                jest.clearAllMocks();
            });
            it('should return correct response', () => {
                expect(response.status).toBe(500);
                expect(response.body).toHaveProperty('message');
                expect(response.body).toHaveProperty('success', false);
            });
        });
    });

    describe('asyncSaveOrSubmitAssignmentResponse', () => {
        const requestPath = `${prefix}/player/2`;
        const ORIGINAL_ENV = process.env;
        const mockSthudentAccessResponse = {
            assignmentId: 1,
            assignmentAssignmentId: '496f4b8f-e34c-49ae-8ad5-81d48d0383f5',
            assignmentOpenDate: null,
            assignmentDueDate: null,
            assignmentAssessmentId: '68b226d0-48c9-4c18-99fe-391d3deb549f',
            assignmentAssignmentStageId: 2,
            assignmentCloseOn: null,
            studentAssignmentId: 1,
            studentAssignmentStartedAt: null,
            studentAssignmentEndedAt: null,
            studentAssignmentPointsScored: 14,
            studentAssignmentTeacherStatusId: 1,
            studentAssignmentStudentStatusId: 1,
            studentAssignmentIsAssignmentLaunched: false,
        };
        beforeEach(async () => {
            jest.resetModules();
            process.env = { ...ORIGINAL_ENV };
        });
        afterEach(async () => {
            process.env = { ...ORIGINAL_ENV };
        });
        describe('when valid input are passed with PLAYER__ENABLE_LAMBDA_FOR_SAVE_SUBMIT=false ', () => {
            beforeEach(async () => {
                mockAxios.post.mockImplementation((urlpath) => {
                    if (urlpath.includes('/assignment-submission/save')) {
                        return Promise.resolve({ data: { success: true } });
                    }
                    return Promise.resolve({ data: { success: false } });
                });
                process.env.PLAYER__ENABLE_LAMBDA_FOR_SAVE_SUBMIT = 'false';
                mockedQueryBuilder.getOne.mockResolvedValueOnce({
                    id: 1,
                    assignmentId: 1,
                });
                mockedQueryBuilder.getRawMany.mockResolvedValueOnce([mockSthudentAccessResponse]);
            });
            afterEach(() => {
                mockAxios.post.mockClear();
                jest.clearAllMocks();
                process.env = { ...ORIGINAL_ENV };
            });
            it('should return correct response', async () => {
                response = await request.post(requestPath).send({ ...saveSubmitInputData() });
                expect(response.status).toBe(200);
                expect(mockAxios.post).toHaveBeenCalled();
                expect(response.body).toHaveProperty('success', true);
            });
        });
        describe('when assignment submission service return error', () => {
            beforeAll(async () => {
                process.env.PLAYER__ENABLE_LAMBDA_FOR_SAVE_SUBMIT = 'false';
                mockAxios.post.mockImplementation((urlpath) => {
                    if (urlpath.includes('/assignment-submission/save')) {
                        return Promise.resolve({ data: { success: false, message: 'error' } });
                    }
                    return Promise.resolve({ data: { success: false } });
                });
                mockedQueryBuilder.getOne.mockResolvedValueOnce({
                    id: 1,
                    assignmentId: 1,
                });
                mockedQueryBuilder.getRawMany.mockResolvedValueOnce([mockSthudentAccessResponse]);
                response = await request.post(requestPath).send({ ...saveSubmitInputData() });
            });
            afterAll(() => {
                process.env = { ...ORIGINAL_ENV };
                mockAxios.post.mockClear();
                jest.clearAllMocks();
            });
            it('should return correct response', () => {
                expect(response.status).toBe(ResponseStatus.INTERNAL_SERVER_ERROR);
                expect(response.body).toHaveProperty('success', false);
                expect(response.body).toHaveProperty('message');
            });
        });
        describe('when invalid input are passed', () => {
            const input = {
                organisationId,
                activityId: 'f86cfbcd-254e-4f3c-add7-d2ab2fffc2e0',
                activityTemplateId: '263d353-ca6c-20-60-1a2c60e21ec3',
                questions: [],
            };
            beforeAll(async () => {
                response = await request.post(requestPath).send({ ...input });
            });
            it('should return correct response', () => {
                expect(response.status).toBe(422);
                expect(response.body).toHaveProperty('success', false);
                expect(response.body).toHaveProperty('message', '"session_id" is not allowed to be empty');
            });
        });
        describe('when error', () => {
            beforeAll(async () => {
                const input = {
                    organisationId,
                    activityId: 'f86cfbcd-254e-4f3c-add7-d2ab2fffc2e0',
                    activityTemplateId: '263d353-ca6c-20-60-1a2c60e21ec3',
                    studentId: 'c96eadb2-8e0b-49e7-ada2-0645734789a7',
                    sessionId: '80c8312d-af7d-48de-8771-5b33ff27632d',
                    questions: [],
                };
                mockedQueryBuilder.getOne.mockImplementationOnce(() => {
                    throw new Error();
                });
                response = await request.post(requestPath).send({ ...input });
            });
            afterAll(async () => {
                jest.clearAllMocks();
                mockedQueryBuilder.getOne.mockReset();
                mockedQueryBuilder.getOne.mockClear();

            });
            it('should return correct response', () => {
                expect(response.status).toBe(500);
                expect(response.body).toHaveProperty('message');
                expect(response.body).toHaveProperty('success', false);
            });
        });
        describe('when valid input are passed with co-presentor true', () => {
            const studentId = 'c96eadb2-8e0b-49e7-ada2-0645734789a7';
            const activityTemplateId = '263d353-ca6c-20-60-1a2c60e21ec3';
            const currentNonce = 101;
            const currentInitTime = 200;
            const nonce = 100;
            const initTime = 100;
            const windowNonce = `${nonce}:${initTime}`;
            beforeEach(async () => {
                redisClient.set(
                    `co-presenter:${activityTemplateId}:${studentId}`, `${currentNonce}:${currentInitTime}`,
                );
                mockAxios.post.mockImplementation((urlpath) => {
                    if (urlpath.includes('/assignment-submission/save')) {
                        return Promise.resolve({ data: { success: true } });
                    }
                    return Promise.resolve({ data: { success: false } });
                });
                process.env.PLAYER__ENABLE_LAMBDA_FOR_SAVE_SUBMIT = 'false';
                mockedQueryBuilder.getOne.mockResolvedValueOnce({
                    id: 1,
                    assignmentId: 1,
                });
                mockedQueryBuilder.getRawMany.mockResolvedValueOnce([mockSthudentAccessResponse]);
            });
            afterEach(() => {
                mockAxios.post.mockClear();
                jest.clearAllMocks();
                process.env = { ...ORIGINAL_ENV };
            });
            it('should return correct response', async () => {
                response = await request.post(requestPath).send({ ...saveSubmitInputData(), isCoPresenter: true  });
                expect(response.status).toBe(200);
                expect(mockAxios.post).toHaveBeenCalled();
                expect(response.body).toHaveProperty('success', true);
            });
            it('should return error for session conflict', async () => {
                response = await request.post(requestPath).send(
                    { ...saveSubmitInputData(), windowNonce, isCoPresenter: true },
                );
                expect(response.status).toBe(ResponseStatus.CONFLICT);
                expect(response.body).toHaveProperty('message');
                expect(response.body).toHaveProperty('success', false);
            });
        });
    });
});
