import { mocked } from 'ts-jest/utils';
import axios from 'axios';
import urlJoin from 'url-join';
import { Lambda, SQS } from 'aws-sdk';
import redisMock from 'redis-mock';
import redis, { RedisClient } from 'redis';
import asyncSaveSubmitHelper from '../asyncSaveSubmitHelper';

const {
    checkLiveClassStatus,
    pushMessageInQueue,
    saveSubmitResponse,
    updateQueueRequestCount,
} = asyncSaveSubmitHelper;

const redisClient: RedisClient = redisMock.createClient();
jest.mock('redis', () => redisMock);
jest.mock('axios');
const mockAxios = mocked(axios, true);
jest.mock('aws-sdk', () => {
    const mockedSQS = {
        sendMessage: jest.fn().mockReturnThis(),
        promise: jest.fn(),
    };
    const mockedLambda = {
        invoke: jest.fn().mockReturnThis(),
        promise: jest.fn(),
    };
    return {
        SQS: jest.fn(() => mockedSQS),
        Lambda: jest.fn(() => mockedLambda),
    };
});

const sqs = new SQS({
    region: 'eu-west-2',
});

const lambda = new Lambda({
    apiVersion: '2015-03-31',
    region: 'eu-west-2',
    endpoint: 'http://localhost:3010',
});

describe('asyncSaveSubmitHelper', () => {
    const checkConsoleError = jest.spyOn(console, 'error');
    let response;
    const activityId = '10000000-0000-4000-A000-000000000001';
    const organisationId = '20000000-0000-4000-A000-000000000001';
    afterAll(async () => {
        redisClient.flushall('ASYNC');
        jest.clearAllMocks();
    });

    describe('checkLiveClassStatus', () => {
        describe('when data is present', () => {
            beforeAll(async () => {
                redisClient.set(`${activityId}:${organisationId}:test1`, `${activityId}-${new Date()}`);
            });

            afterAll(async () => {
                redisClient.flushall('ASYNC');
            });

            it('Get data for all students for given assignmentId', async () => {
                response = await checkLiveClassStatus(activityId, organisationId);
                expect(response).toEqual(true);
            });
        });
        describe('when no data is present', () => {
            beforeAll(async () => {
                redisClient.flushall('ASYNC');
            });

            afterAll(async () => {
                redisClient.flushall('ASYNC');
            });

            it('checks if live class is active for given activity', async () => {
                response = await checkLiveClassStatus(activityId, organisationId);
                expect(response).toEqual(false);
            });
        });
    });

    describe('saveSubmitResponse', () => {
        let mockPost;

        describe('when valid', () => {
            beforeAll(async () => {
                mockPost = mockAxios.post.mockImplementationOnce(() =>
                    Promise.resolve({ data: { success: true } }),
                );
            });
            it('should make axios call to assignment submission service with correct data', async () => {
                const data = {
                    activityId,
                    organisationId,
                    studentHasAccess: {},
                    questions: [],
                    isSubmitOrSave: 2,
                    studentId: '00000000-0000-4000-A000-000000000001',
                    requestId: '00000000-0000-4000-A000-000000000001',
                };
                const axiosUrl = urlJoin(process.env.ASSIGNMENT_SUBMISSION_SERVICE, '/assignment-submission/save');
                response = await saveSubmitResponse(data, data.requestId);
                expect(mockPost).toHaveBeenCalledWith(
                    axiosUrl,
                    { data },
                    expect.any(Object),
                );
                expect(response).toHaveProperty('success', true);
                expect(response).toHaveProperty('message', expect.any(String));
            });
        });

        describe('when valid but no request Id', () => {
            beforeAll(async () => {
                mockPost = mockAxios.post.mockImplementationOnce(() =>
                    Promise.resolve({ data: { success: true } }),
                );
            });
            it('should make axios call to assignment submission service with correct data', async () => {
                const data = {
                    activityId,
                    organisationId,
                    studentHasAccess: {},
                    questions: [],
                    isSubmitOrSave: 1,
                    studentId: '00000000-0000-4000-A000-000000000001',
                    requestId: '00000000-0000-4000-A000-000000000001',
                };
                const axiosUrl = urlJoin(process.env.ASSIGNMENT_SUBMISSION_SERVICE, '/assignment-submission/save');
                response = await saveSubmitResponse(data, '');
                expect(mockPost).toHaveBeenCalledWith(
                    axiosUrl,
                    { data },
                    expect.any(Object),
                );
                expect(response).toHaveProperty('success', true);
                expect(response).toHaveProperty('message', expect.any(String));
            });
        });

        describe('when Error', () => {
            beforeAll(async () => {
                mockPost = mockAxios.post.mockImplementationOnce(() => {
                    throw new Error();
                });
            });
            it('should console log error', async () => {
                const data = {
                    activityId,
                    organisationId,
                    studentHasAccess: {},
                    questions: [],
                    isSubmitOrSave: 2,
                    studentId: '00000000-0000-4000-A000-000000000001',
                    requestId: '00000000-0000-4000-A000-000000000001',
                };
                const axiosUrl = urlJoin(process.env.ASSIGNMENT_SUBMISSION_SERVICE, '/assignment-submission/save');
                expect.assertions(3);
                try {
                    response = await saveSubmitResponse(data, data.requestId);
                } catch (error) {
                    expect(error).toBeDefined();
                    expect(checkConsoleError).toBeCalled();
                    expect(mockPost).toHaveBeenCalledWith(
                        axiosUrl,
                        { data },
                        expect.any(Object),
                    );
                }
            });
        });
    });

    describe('pushMessageInQueue', () => {
        const ORIGINAL_ENV = process.env;
        const message = {
            activityId,
            organisationId,
            studentHasAccess: {},
            questions: [],
            isSubmitOrSave: 2,
            studentId: '00000000-0000-4000-A000-000000000001',
            requestId: '00000000-0000-4000-A000-000000000001',
        };
        const studentAssignmentId = 100;
        const assignmentId = activityId;
        describe('when request is for simple queue', () => {
            const isSimpleQueueRequest = true;
            describe('when env stage is local', () => {
                beforeAll(async () => {
                    jest.resetModules();
                    process.env = { ...ORIGINAL_ENV };
                    (lambda.invoke().promise as jest.MockedFunction<any>).mockReset();
                });
                afterAll(() => {
                    (lambda.invoke().promise as jest.MockedFunction<any>).mockReset();
                    jest.clearAllMocks();
                });
                it('should make dummyQueue request', async () => {
                    process.env.STAGE = 'local';
                    (lambda.invoke().promise as jest.MockedFunction<any>).mockResolvedValueOnce('mocked data');
                    await pushMessageInQueue(isSimpleQueueRequest, message, studentAssignmentId, assignmentId);
                    const lambdaData = {
                        FunctionName: 'itemengine-player-api-local-save-student-attempts-simple',
                        InvocationType: 'RequestResponse',
                        Payload: JSON.stringify({ data: message }),
                    };
                    expect(lambda.invoke).toBeCalledWith(lambdaData);
                    expect(lambda.invoke().promise).toBeCalled();
                });
            });
            describe('when env stage is not local', () => {
                beforeAll(async () => {
                    jest.resetModules();
                    process.env = { ...ORIGINAL_ENV };
                    (sqs.sendMessage().promise as jest.MockedFunction<any>).mockReset();
                });
                afterAll(() => {
                    process.env = { ...ORIGINAL_ENV };
                    (sqs.sendMessage().promise as jest.MockedFunction<any>).mockReset();
                    jest.clearAllMocks();
                });
                it('should make simple queue sqs request', async () => {
                    process.env.STAGE = 'qa';
                    (sqs.sendMessage().promise as jest.MockedFunction<any>).mockResolvedValueOnce('mocked data');
                    await pushMessageInQueue(isSimpleQueueRequest, message, studentAssignmentId, assignmentId);
                    const sqsData = {
                        MessageBody: JSON.stringify(message),
                        QueueUrl: process.env.SAVE_STUDENT_ATTEMPTS_SIMPLE_SQS_URL,
                        MessageGroupId: `${studentAssignmentId}`,
                    };
                    expect(sqs.sendMessage).toBeCalledWith(sqsData);
                    expect(sqs.sendMessage().promise).toBeCalled();
                });
            });
        });
        describe('when request is for priority queue', () => {
            const isSimpleQueueRequest = false;
            describe('when env stage is local', () => {
                beforeAll(async () => {
                    jest.resetModules();
                    process.env = { ...ORIGINAL_ENV };
                    (lambda.invoke().promise as jest.MockedFunction<any>).mockReset();
                });
                afterAll(() => {
                    process.env = { ...ORIGINAL_ENV };
                    (lambda.invoke().promise as jest.MockedFunction<any>).mockReset();
                    jest.clearAllMocks();
                });
                it('should make dummyQueue request', async () => {
                    process.env.STAGE = 'local';
                    (lambda.invoke().promise as jest.MockedFunction<any>).mockResolvedValueOnce('mocked data');
                    await pushMessageInQueue(isSimpleQueueRequest, message, studentAssignmentId, assignmentId);
                    const lambdaData = {
                        FunctionName: 'itemengine-player-api-local-save-student-attempts-priority',
                        InvocationType: 'RequestResponse',
                        Payload: JSON.stringify({ data: message }),
                    };
                    expect(lambda.invoke).toBeCalledWith(lambdaData);
                    expect(lambda.invoke().promise).toBeCalled();
                });
            });
            describe('when env stage is not local', () => {
                beforeAll(async () => {
                    jest.resetModules();
                    process.env = { ...ORIGINAL_ENV };
                    (sqs.sendMessage().promise as jest.MockedFunction<any>).mockReset();
                });
                afterAll(() => {
                    process.env = { ...ORIGINAL_ENV };
                    (sqs.sendMessage().promise as jest.MockedFunction<any>).mockReset();
                    jest.clearAllMocks();
                });
                it('should make priority queue sqs request', async () => {
                    process.env.STAGE = 'qa';
                    (sqs.sendMessage().promise as jest.MockedFunction<any>).mockResolvedValueOnce('mocked data');
                    await pushMessageInQueue(isSimpleQueueRequest, message, studentAssignmentId, assignmentId);
                    const sqsData = {
                        MessageBody: JSON.stringify(message),
                        QueueUrl: process.env.SAVE_STUDENT_ATTEMPTS_PRIORITY_SQS_URL,
                        MessageGroupId: `${studentAssignmentId}`,
                    };
                    expect(sqs.sendMessage).toBeCalledWith(sqsData);
                    expect(sqs.sendMessage().promise).toBeCalled();
                });
            });
        });
    });

    describe('updateQueueRequestCount', () => {
        const studentAssignmentId = '10000000-0000-4000-A000-000000000001';
        const assignmentId = '20000000-0000-4000-A000-000000000001';
        const isLiveClassEnabled = false;
        const val = {
            requestCount: {
                simple: 1,
                priority: 0,
            },
        };
        describe('when data is present', () => {
            beforeAll(async () => {
                jest.clearAllMocks();
                redisClient.watch = redisClient.get;
                jest.spyOn(redis, 'createClient').mockReturnValue(redisClient);
                redisClient.set(`${studentAssignmentId}:${assignmentId}`, JSON.stringify(val));
            });

            afterAll(async () => {
                redisClient.flushall('ASYNC');
            });

            it('Get data for all students for given assignmentId', async () => {
                response = await updateQueueRequestCount(studentAssignmentId, assignmentId, isLiveClassEnabled);
                expect(response).toEqual(true);
            });

            it('when isLiveClassEnabled is true', async () => {
                response = await updateQueueRequestCount(studentAssignmentId, assignmentId, true);
                expect(response).toEqual(true);
            });
        });

        describe('when data is not present', () => {
            beforeAll(async () => {
                jest.clearAllMocks();
                redisClient.watch = redisClient.get;
                jest.spyOn(redis, 'createClient').mockReturnValue(redisClient);
                redisClient.flushall('ASYNC');
            });

            afterAll(async () => {
                redisClient.flushall('ASYNC');
            });

            it('when to key cache and liveClass enabled', async () => {
                response = await updateQueueRequestCount(studentAssignmentId, assignmentId, true);
                expect(response).toEqual(false);
            });

            it('when to key cache and liveClass enabled false', async () => {
                response = await updateQueueRequestCount(studentAssignmentId, assignmentId, isLiveClassEnabled);
                expect(response).toEqual(false);
            });
        });
        describe('when Error', () => {
            beforeAll(async () => {
                jest.clearAllMocks();
                redisClient.watch = redisClient.get;
                redisClient.get = jest.fn().mockImplementation(() => {
                    throw new Error();
                });
                jest.spyOn(redis, 'createClient').mockReturnValue(redisClient);
                redisClient.flushall('ASYNC');
            });

            afterAll(async () => {
                redisClient.flushall('ASYNC');
            });

            it('should log error', async () => {
                response = await updateQueueRequestCount(studentAssignmentId, assignmentId, isLiveClassEnabled);
                expect(checkConsoleError).toBeCalled();
            });
        });
    });
});
