import {
    mockedQueryBuilder,
    mockedRepository,
    connectionObj as mockConnection,
} from '../../__mocks__/typeorm.mock';
import { mocked } from 'ts-jest/utils';
import axios from 'axios';
import urlJoin from 'url-join';
import redisMock from 'redis-mock';
import mockData from '../../__mocks__/mockData';
import playerControllerHelpers from '../playerControllerHelpers';
import constants from './../../../../utils/constants';
import { TimeLogger } from '../../../../utils/timeLogger';

jest.mock('axios');
const mockAxios = mocked(axios, true);

const redisClient = redisMock.createClient();
jest.mock('../../../../redisManager', () => ({
    getRedisInstance: jest.fn(() => redisClient)}),
);

const {
    questionQuestionpartMapMockData,
    assignmentsMockData,
    assignmentStudentMapMockData,
} = mockData;

const {
    studentPlayerViewSchema,
    saveSubmitInputSchema,
    checkStudentExists,
    storeStudentData,
    getAssignmentDataByReferenceId,
    fetchAssignmentStudentResponse,
    doesStudentHasAccessOrIsAuthorized,
    fetchAssessmentCorrectAnswerView,
    insertResponses,
    getResponseArray,
    getPointsScoredAndIsGraded,
    getActivityData,
    doesStudentHaveAccessToAssignment,
    storeAssignmentQuestionMaps,
    extractQuestionPartsMap,
    fetchAssignmentQuestionMap,
} = playerControllerHelpers;

const {
    submitOrSave: { SAVE },
} = constants;

describe('playerControllerHelper', () => {
    let response;
    const checkConsoleError = jest.spyOn(console, 'error');
    beforeAll(async () => {
        jest.clearAllMocks();
    });

    afterAll(async () => {
        jest.clearAllMocks();
        mockedQueryBuilder.getRawOne.mockClear();
        mockedRepository.save.mockClear();
        mockedQueryBuilder.getOne.mockClear();
        mockedQueryBuilder.getRawMany.mockClear();
        mockedQueryBuilder.execute.mockClear();
        mockedQueryBuilder.getMany.mockClear();
    });

    describe('studentPlayerViewSchema', () => {
        it('when valid input', async () => {
            const input = {
                organisation_id: '00000000-0000-4000-A000-000000000001',
                student_id: '00000000-0000-4000-A000-000000000001',
                activity_id: '00000000-0000-4000-A000-000000000001',
                activity_template_id: '00000000-0000-4000-A000-000000000001',
            };
            const { error, value } = await studentPlayerViewSchema(input);
            expect(error).toBeUndefined();
            expect(value).toBeDefined();
        });
        it('should return error for invalid input', async () => {
            const input = {
                session_id: '00000000-0000-4000-A000-000000000001',
                student_id: '00000000-0000-4000-A000-000000000001',
                activity_id: '00000000-0000-4000-A000-000000000001',
                activity_template_id: '00000000-0000-4000-A000-000000000001',
            };
            const { error, value } = await studentPlayerViewSchema(input);
            expect(error).toBeDefined();
            expect(value).toBeDefined();
        });
    });

    describe('saveSubmitInputSchema', () => {
        it('when valid input', async () => {
            const input = {
                organisation_id: '00000000-0000-4000-A000-000000000001',
                student_id: '00000000-0000-4000-A000-000000000001',
                activity_id: '00000000-0000-4000-A000-000000000001',
                activity_template_id: '00000000-0000-4000-A000-000000000001',
                questions: {
                    question_id: '00000000-0000-4000-A000-000000000001',
                    question_parts: '00000000-0000-4000-A000-000000000001',
                },
            };
            const { error, value } = await saveSubmitInputSchema(input);
            expect(error).toBeDefined();
            expect(value).toBeDefined();
        });

        it('should return error for invalid input', async () => {
            const input = {
                session_id: '00000000-0000-4000-A000-000000000001',
                student_id: '00000000-0000-4000-A000-000000000001',
                activity_id: '00000000-0000-4000-A000-000000000001',
                activity_template_id: '00000000-0000-4000-A000-000000000001',
                questions: {
                    question_id: '00000000-0000-4000-A000-000000000001',
                    question_parts: '00000000-0000-4000-A000-000000000001',
                },
            };
            const { error, value } = await saveSubmitInputSchema(input);
            expect(error).toBeDefined();
            expect(value).toBeDefined();
        });
    });

    describe('checkStudentExists', () => {
        beforeAll(async () => {
            mockedQueryBuilder.getRawOne.mockResolvedValueOnce(assignmentStudentMapMockData[0])
            .mockResolvedValueOnce(assignmentStudentMapMockData[0]);
        });
        afterAll(async () => {
            jest.clearAllMocks();
            mockedQueryBuilder.getRawOne.mockReset();
            mockedQueryBuilder.getRawOne.mockClear();
        });
        it('should return if student is exists', async () => {
            const params = {
                conn: mockConnection,
                sessionId: '00000000-0000-0000-00000-000000000001',
                studentId: 'de006704-0000-1121-bb37-024200130002',
                assessmentId: '00000000-0000-4000-A000-000000000001',
                activityId: '99000000-0000-4000-A000-000000000001',
            };
            response = await checkStudentExists(params);
            expect(response).toHaveProperty('id');
        });

        it('when organisation Id passed', async () => {
            const params = {
                conn: mockConnection,
                sessionId: '00000000-0000-0000-00000-000000000001',
                studentId: 'de006704-0000-1121-bb37-024200130002',
                assessmentId: '00000000-0000-4000-A000-000000000001',
                activityId: '99000000-0000-4000-A000-000000000001',
                organisationId: '99000000-0000-4000-A000-000000000001',
            };
            response = await checkStudentExists(params);
            expect(response).toHaveProperty('id');
        });
    });

    describe('storeStudentData', () => {
        beforeAll(async () => {
            mockedRepository.save.mockResolvedValueOnce(assignmentStudentMapMockData[0]);
        });
        afterAll(async () => {
            jest.clearAllMocks();
            mockedRepository.save.mockReset();
            mockedRepository.save.mockClear();
        });
        it('should return store student data', async () => {
            const params = {
                transaction: mockConnection,
                assignmentId: 1,
                studentId: 'de006704-0000-1121-bb37-024200130002',
                sessionId: '00a6fa25-df29-4701-9077-557932591766',
            };
            response = await storeStudentData(params);
            expect(response).toHaveProperty('id');
            expect(response).toHaveProperty('assignmentId');
            expect(response).toHaveProperty('gradingStatusId');
            expect(response).toHaveProperty('teacherStatusId');
            expect(response).toHaveProperty('studentStatusId');
            expect(response).toHaveProperty('studentId');
            expect(response).toHaveProperty('sessionId');
        });
    });

    describe('getAssignmentDataByReferenceId', () => {
        beforeAll(async () => {
            mockedQueryBuilder.getOne.mockResolvedValueOnce(assignmentsMockData[0])
            .mockResolvedValueOnce(assignmentsMockData[0])
            .mockResolvedValueOnce(assignmentsMockData[0]);
        });
        afterAll(async () => {
            jest.clearAllMocks();
            mockedQueryBuilder.getOne.mockReset();
            mockedQueryBuilder.getOne.mockClear();
        });
        it('should return assignment data by referenceId', async () => {
            const params = {
                conn: mockConnection,
                assignmentReferenceId: '99000000-0000-4000-A000-000000000001',
                organisationId: '99000000-0000-4000-A000-000000000001',
                selectFields: ['assignment.id'],
            };
            response = await getAssignmentDataByReferenceId(params);
            expect(response).toHaveProperty('id');
        });

        it('if no organisationId passed', async () => {
            const params = {
                conn: mockConnection,
                assignmentReferenceId: '99000000-0000-4000-A000-000000000001',
                selectFields: ['assignment.id'],
            };
            response = await getAssignmentDataByReferenceId(params);
            expect(response).toHaveProperty('id');
        });

        it('if no selectFields passed', async () => {
            const params = {
                conn: mockConnection,
                assignmentReferenceId: '99000000-0000-4000-A000-000000000001',
                organisationId: '99000000-0000-4000-A000-000000000001',
            };
            response = await getAssignmentDataByReferenceId(params);
            expect(response).toHaveProperty('id');
            expect(response).toHaveProperty('assignmentId');
        });
    });

    describe('fetchAssignmentStudentResponse', () => {
        beforeAll(async () => {
            mockedQueryBuilder.getRawMany.mockResolvedValueOnce([questionQuestionpartMapMockData[0]]);
        });
        afterAll(async () => {
            jest.clearAllMocks();
            mockedQueryBuilder.getRawMany.mockReset();
            mockedQueryBuilder.getRawMany.mockClear();
        });
        it('should return assignment student response', async () => {
            const params = {
                conn: mockConnection,
                sessionId: '00000000-0000-0000-00000-000000000001',
                studentId: 'de006704-0000-1121-bb37-024200130002',
                assessmentId: '00000000-0000-4000-A000-000000000001',
                activityId: '99000000-0000-4000-A000-000000000001',
            };
            response = await fetchAssignmentStudentResponse(params);
            response.forEach((res) => {
                expect(res).toHaveProperty('id');
                expect(res).toHaveProperty('question_part_id');
                expect(res).toHaveProperty('assignmentQuestionMap');
                expect(res).toHaveProperty('response');
            });
        });
    });

    describe('fetchAssessmentCorrectAnswerView', () => {
        let mockPost;
        const activityMockedData = { id: 1 };
        const url = urlJoin(process.env.ACTIVITY_SERVICE, '/activity/view/correctAnswer');
        beforeAll(async () => {
            mockPost = mockAxios.post.mockImplementationOnce(() =>
                Promise.resolve({ data: { activityData: activityMockedData } }),
            );
            const assessmentParams = { activityTemplateId: '263d353-ca6c-20-60-1a2c60e21ec3' };
            response = await fetchAssessmentCorrectAnswerView(assessmentParams);
        });

        afterAll(() => {
            jest.clearAllMocks();
        });

        it('should call axios get with correct parameters', () => {
            expect(mockPost).toHaveBeenCalledWith(
                url,
                { activityReference: '263d353-ca6c-20-60-1a2c60e21ec3' },
                expect.any(Object),
            );
        });

        it('should return correct data', () => {
            expect(response).toEqual(activityMockedData);
        });
    });

    describe('doesStudentHasAccessOrIsAuthorized', () => {
        beforeAll(async () => {
            const mockResponse = {
                assignmentId: 1,
                assignmentAssignmentId: '496f4b8f-e34c-49ae-8ad5-81d48d0383f5',
                assignmentOpenDate: null,
                assignmentDueDate: null,
                assignmentAssessmentId: '68b226d0-48c9-4c18-99fe-391d3deb549f',
                assignmentAssignmentStageId: 2,
                assignmentCloseOn: null,
                studentAssignmentId: 1,
                studentAssignmentStartedAt: null,
                studentAssignmentEndedAt: null,
                studentAssignmentPointsScored: 14,
                studentAssignmentTeacherStatusId: 1,
                studentAssignmentStudentStatusId: 1,
                studentAssignmentIsAssignmentLaunched: false,
            };
            mockedQueryBuilder.getRawMany.mockResolvedValueOnce([mockResponse]);
        });
        afterAll(async () => {
            jest.clearAllMocks();
            mockedQueryBuilder.getRawMany.mockReset();
            mockedQueryBuilder.getRawMany.mockClear();
        });

        it('should return correct data', async() => {
            const params = {
                studentId: 'c96eadb2-8e0b-49e7-ada2-0645734789a7',
                assignmentId: '496f4b8f-e34c-49ae-8ad5-81d48d0383f5',
            };
            response = await doesStudentHasAccessOrIsAuthorized(params);
            expect(response).toHaveProperty('assignmentId');
            expect(response).toHaveProperty('assignmentAssignmentId');
            expect(response).toHaveProperty('studentAssignmentId');
            expect(response).toHaveProperty('assignmentAssignmentStageId');
            expect(response).toHaveProperty('studentAssignmentPointsScored');
        });
    });

    describe('doesStudentHaveAccessToAssignment', () => {
        beforeAll(async () => {
            const mockResponse = {
                assignmentId: 1,
                assignmentAssignmentId: '496f4b8f-e34c-49ae-8ad5-81d48d0383f5',
                assignmentOpenDate: null,
                assignmentDueDate: null,
                assignmentAssessmentId: '68b226d0-48c9-4c18-99fe-391d3deb549f',
                assignmentAssignmentStageId: 2,
                assignmentCloseOn: null,
                studentAssignmentId: 1,
                studentAssignmentStartedAt: null,
                studentAssignmentEndedAt: null,
                studentAssignmentPointsScored: 14,
                studentAssignmentTeacherStatusId: 1,
                studentAssignmentStudentStatusId: 1,
                studentAssignmentIsAssignmentLaunched: false,
            };
            mockedQueryBuilder.getRawMany.mockResolvedValueOnce([mockResponse])
            .mockResolvedValueOnce([])
            .mockImplementationOnce(() => {
                throw new Error();
            });
        });
        afterAll(async () => {
            jest.clearAllMocks();
            mockedQueryBuilder.getRawMany.mockReset();
            mockedQueryBuilder.getRawMany.mockClear();
        });

        it('should return correct data', async() => {
            const params = {
                studentId: 'c96eadb2-8e0b-49e7-ada2-0645734789a7',
                assignmentId: '496f4b8f-e34c-49ae-8ad5-81d48d0383f5',
            };
            response = await doesStudentHasAccessOrIsAuthorized(params);
            expect(response).toHaveProperty('assignmentId');
            expect(response).toHaveProperty('assignmentAssignmentId');
            expect(response).toHaveProperty('studentAssignmentId');
            expect(response).toHaveProperty('assignmentAssignmentStageId');
            expect(response).toHaveProperty('studentAssignmentPointsScored');
        });

        it('should return null for no result found', async() => {
            const params = {
                studentId: 'c96eadb2-8e0b-49e7-ada2-0645734789a7',
                assignmentId: '496f4b8f-e34c-49ae-8ad5-81d48d0383f5',
            };
            response = await doesStudentHaveAccessToAssignment(params);
            expect(response).toEqual(null);
        });

        it('should return null for invalid input', async() => {
            const params = {
                studentId: '',
                assignmentId: '496f4b8f-e34c-49ae-8ad5-81d48d0383f5',
            };
            response = await doesStudentHaveAccessToAssignment(params);
            expect(response).toEqual(null);
        });
    });

    describe('insertResponses', () => {
        beforeAll(async () => {
            mockedRepository.save.mockResolvedValueOnce({
                questionParts: [{
                    id: 1,
                }],
            }).mockResolvedValueOnce({ id: 2 })
            .mockImplementationOnce(() => {
                throw { code: 23505 };
            }).mockImplementationOnce(() => {
                throw new Error();
            });
        });
        afterAll(async () => {
            jest.clearAllMocks();
            mockedRepository.save.mockReset();
            mockedRepository.save.mockClear();
        });
        const NEW_QUESTION = 1;
        const NEW_QUESTION_PART = 2;
        it('when NEW_QUESTION', async () => {
            const insertData = {
                assignmentStudentId: 1,
                responseTextsArray: [
                    {
                        response: [
                            { xmlIdentifier: 'RESPONSE', responseText: 'ChoiceA' },
                            { xmlIdentifier: 'RESPONSE', responseText: 'ChoiceD' },
                        ],
                    },
                ],
                response: [
                    { xmlIdentifier: 'RESPONSE', responseText: 'ChoiceA' },
                    { xmlIdentifier: 'RESPONSE', responseText: 'ChoiceD' },
                ],
                isSubmitOrSave: null,
                isEmptyMultipleChoiceResponse: false,
                foreignKey: 1,
                newItemToSubmit: NEW_QUESTION,
            };
            response = await insertResponses(insertData);
            expect(response).toEqual(1);
        });

        it('when NEW_QUESTION_PART', async () => {
            const insertData = {
                assignmentStudentId: 1,
                responseTextsArray: [
                    {
                        response: [
                            { xmlIdentifier: 'RESPONSE', responseText: 'ChoiceA' },
                            { xmlIdentifier: 'RESPONSE', responseText: 'ChoiceD' },
                        ],
                    },
                ],
                response: [
                    { xmlIdentifier: 'RESPONSE', responseText: 'ChoiceA' },
                    { xmlIdentifier: 'RESPONSE', responseText: 'ChoiceD' },
                ],
                isSubmitOrSave: null,
                isEmptyMultipleChoiceResponse: false,
                foreignKey: 1,
                newItemToSubmit: NEW_QUESTION_PART,
            };
            response = await insertResponses(insertData);
            expect(response).toEqual(2);
        });

        it('when Error with code', async () => {
            const insertData = {
                assignmentStudentId: 1,
                responseTextsArray: [
                    {
                        response: [
                            { xmlIdentifier: 'RESPONSE', responseText: 'ChoiceA' },
                            { xmlIdentifier: 'RESPONSE', responseText: 'ChoiceD' },
                        ],
                    },
                ],
                response: [
                    { xmlIdentifier: 'RESPONSE', responseText: 'ChoiceA' },
                    { xmlIdentifier: 'RESPONSE', responseText: 'ChoiceD' },
                ],
                isSubmitOrSave: null,
                isEmptyMultipleChoiceResponse: false,
                foreignKey: 1,
                newItemToSubmit: NEW_QUESTION_PART,
            };
            response = await insertResponses(insertData);
            expect(checkConsoleError).toHaveBeenCalled();
            expect(response).toEqual(null);
        });

        it('when !item || !foreignKey', async () => {
            const insertData = {
                assignmentStudentId: 1,
                responseTextsArray: [
                    {
                        response: [
                            { xmlIdentifier: 'RESPONSE', responseText: 'ChoiceA' },
                            { xmlIdentifier: 'RESPONSE', responseText: 'ChoiceD' },
                        ],
                    },
                ],
                response: [],
                isSubmitOrSave: null,
                isEmptyMultipleChoiceResponse: false,
                foreignKey: null,
                newItemToSubmit: NEW_QUESTION_PART,
            };
            response = await insertResponses(insertData);
            expect(response).toEqual(null);
        });

        it('when no responsetext and is SAVE', async () => {
            const insertData = {
                assignmentStudentId: 1,
                responseTextsArray: [
                    {
                        response: [
                            { xmlIdentifier: 'RESPONSE', responseText: 'ChoiceA' },
                            { xmlIdentifier: 'RESPONSE', responseText: '' },
                        ],
                    },
                ],
                response: [],
                isSubmitOrSave: SAVE,
                isEmptyMultipleChoiceResponse: true,
                foreignKey: 1,
                newItemToSubmit: NEW_QUESTION_PART,
            };
            response = await insertResponses(insertData);
            expect(response).toBeUndefined();
        });
    });

    describe('getResponseArray', () => {
        it('should return reponse array', () => {
            const params = {
                responseTexts: [
                    { xmlIdentifier: 'RESPONSE', responseText: 'ChoiceA' },
                    { xmlIdentifier: 'RESPONSE', responseText: 'ChoiceD' },
                ],
            };
            response = getResponseArray(params);
            response.forEach((res) => {
                expect(res).toHaveProperty('xmlIdentifier');
                expect(res).toHaveProperty('responseText');
            });
        });
        it('when no responseText', () => {
            const params = {
                responseTexts: [
                    { xmlIdentifier: 'RESPONSE', responseText: null },
                    { xmlIdentifier: 'RESPONSE', responseText: null },
                ],
            };
            response = getResponseArray(params);
            expect(response[0]).toEqual(null);
        });
    });

    describe('getPointsScoredAndIsGraded', () => {
        it('should return isCompletelyGraded false for null input', async () => {
            const { isCompletelyGraded, pointsScored } = getPointsScoredAndIsGraded(null);
            expect(pointsScored).toEqual(null);
            expect(isCompletelyGraded).toEqual(false);
        });
        it('should return isCompletelyGraded true for 0', async () => {
            const { isCompletelyGraded, pointsScored } = getPointsScoredAndIsGraded(0);
            expect(pointsScored).toEqual(0);
            expect(isCompletelyGraded).toEqual(true);
        });
        it('should return isCompletelyGraded true for valid input', async () => {
            const { isCompletelyGraded, pointsScored } = getPointsScoredAndIsGraded(10);
            expect(pointsScored).toEqual(10);
            expect(isCompletelyGraded).toEqual(true);
        });
    });

    describe('extractQuestionPartsMap', () => {
        const input = [
            {
                questionId: 1,
                questionParts: [
                    { questionPartId: 1 },
                ],
            },
        ];
        it('should return correct value', async () => {
            const response = extractQuestionPartsMap(input);
            response.forEach((res) => {
                expect(res).toHaveProperty('questionId');
                expect(res).toHaveProperty('questionPartId');
            });
        });
    });

    describe('storeAssignmentQuestionMaps', () => {
        beforeEach(() => {
            mockedRepository.save.mockResolvedValueOnce({});
            mockedQueryBuilder.execute.mockResolvedValueOnce({});
        });
        afterEach(() => {
            jest.clearAllMocks();
            mockedRepository.save.mockReset();
            mockedRepository.save.mockClear();
            mockedQueryBuilder.execute.mockReset();
            mockedQueryBuilder.execute.mockClear();
        });

        const input = {
            conn: mockConnection,
            studentAssignmentId: '10000000-0000-4000-A000-000000001112',
            items: [
                {
                    questionId: 1,
                    questions: [
                        { id: 1 },
                    ],
                },
            ],
            assignmentQuestionMaps: [
                {
                    questionId: 1,
                    id: 1,
                    questionParts: [
                        { questionPartId: 1 },
                    ],
                },
            ],
            questionQuestionPartMaps: [
                { assignmentQuestionMappingId: 1, questionPartId: 1 },
            ],
            sessionId: '10000000-0000-4000-A000-000000001111',
        };
        it('when valid input and itemAssignmentQuestion found', async () => {
            const response = await storeAssignmentQuestionMaps(input);
            response.forEach((res) => {
                expect(res).toHaveProperty('questionId');
                expect(res).toHaveProperty('questionPartId');
            });
        });

        it('when valid input and itemAssignmentQuestion not found', async () => {
            const response = await storeAssignmentQuestionMaps({
                ...input,
                items: [
                    {
                        questionId: 2,
                        questions: [
                            { id: 1 },
                        ],
                    },
                ],
            });
            response.forEach((res) => {
                expect(res).toHaveProperty('questionId');
                expect(res).toHaveProperty('questionPartId');
            });
        });

        it('when valid input and new question found', async () => {
            const response = await storeAssignmentQuestionMaps({
                ...input,
                items: [
                    {
                        questionId: 1,
                        questions: [
                            { id: 1 },
                            { id: 2 },
                        ],
                    },
                ],
            });
            response.forEach((res) => {
                expect(res).toHaveProperty('questionId');
                expect(res).toHaveProperty('questionPartId');
            });
        });

        it('when valid input and deleted  question found', async () => {
            const response = await storeAssignmentQuestionMaps({
                ...input,
                questionQuestionPartMaps: [
                    { assignmentQuestionMappingId: 1, questionPartId: 1 },
                    { assignmentQuestionMappingId: 1, questionPartId: 2 },
                ],
            });
            response.forEach((res) => {
                expect(res).toHaveProperty('questionId');
                expect(res).toHaveProperty('questionPartId');
            });
        });
    });

    describe('fetchAssignmentQuestionMap', () => {
        beforeEach(async () => {
            mockedQueryBuilder.getMany.mockResolvedValueOnce([{
                qustion_part_id: 10,
                response_id: 11,
            }]);
        });
        afterEach(async () => {
            jest.clearAllMocks();
            mockedQueryBuilder.getMany.mockReset();
            mockedQueryBuilder.getMany.mockClear();
        });

        it('should return correct data', async () => {
            response = await fetchAssignmentQuestionMap({ conn: mockConnection });
            response.forEach((res) => {
                expect(res).toHaveProperty('qustionPartId');
                expect(res).toHaveProperty('responseId');
            });
        });
    });

    describe('getActivityData', () => {
        const activityData = { id: 1 };
        const activityDataPath = urlJoin(process.env.ACTIVITY_SERVICE, '/activity/view/correctAnswer');
        let mockPost;
        const timeLogger = new TimeLogger('getActivityData-test', []);
        describe('getActivityData by axios call', () => {
            beforeAll(async () => {
                await redisClient.flushall();
                mockPost = mockAxios.post.mockImplementation((urlpath) => {
                    if (urlpath.includes('/activity/view/correctAnswer')) {
                        return Promise.resolve({ data: { activityData } });
                    }
                    return Promise.resolve({ data: { success: true } });
                });
                const params = {
                    timeLogger,
                    activityTemplateId: activityData.id,
                    organisationId: 'mockid',
                };
                response = await getActivityData(params);
            });

            afterAll(() => {
                jest.clearAllMocks();
            });

            it('should call remote get apis with correct parameters', () => {
                const firstGetRequest = mockPost.mock.calls[0];
                expect(firstGetRequest[0]).toEqual(activityDataPath);
                expect(firstGetRequest[1]).toMatchObject(expect.any(Object));
            });

            it('should return correct data', () => {
                expect(response).toEqual(activityData);
            });
        });
        describe('getActivityData by redis cache', () => {
            beforeAll(async () => {
                const params = {
                    timeLogger,
                    activityTemplateId: activityData.id,
                    organisationId: 'mockid',
                };
                redisClient.set(
                    `fetchAssessmentCorrectAnswerView:${params.activityTemplateId}:${params.organisationId}`,
                    JSON.stringify(activityData),
                );
                response = await getActivityData(params);
            });

            afterAll(() => {
                jest.clearAllMocks();
            });
            it('should return correct data', () => {
                expect(response).toEqual(activityData);
            });
        });

        describe('when Error', () => {
            beforeAll(async () => {
                await redisClient.flushall();
                mockPost = mockAxios.post.mockImplementation((urlpath) => {
                    if (urlpath.includes('/activity/view/correctAnswer')) {
                        throw new Error();
                    }
                    return Promise.resolve({ data: { success: true } });
                });
                const params = {
                    timeLogger,
                    activityTemplateId: activityData.id,
                    organisationId: 'mockid',
                };
                response = await getActivityData(params);
            });

            afterAll(() => {
                jest.clearAllMocks();
            });
            it('should call console error', () => {
                expect(checkConsoleError).toBeCalled();
                expect(response).toEqual(null);
            });
        });
    });
});
