import { log } from '../../../utils/LogError';
import { getRedisInstance } from '../../../redisManager';
import { promisify } from 'util';
import { SQS } from 'aws-sdk';
import axios from 'axios';
import https from 'https';
import { v4 as uuidv4 } from 'uuid';
import urlJoin from 'url-join';
import dummyQueue from '../../../lambda/dummyQueue';
import constants from '../../../utils/constants';

const {
    submitOrSave: { SAVE },
} = constants;

const checkLiveClassStatus = async(activityId, organisationId) => {
    let redisClient;
    try {
        redisClient = getRedisInstance();
        const asyncKeys = promisify(redisClient.keys).bind(redisClient);

        const matchingKeys = await asyncKeys(`${activityId}:${organisationId}:*`);
        return matchingKeys?.length > 0;
    } catch (error) {
        console.error(error);
        return false;
    } finally {
        redisClient?.quit();
    }
};

const updateQueueRequestCount = async(studentAssignmentId, assignmentId, isLiveClassEnabled) => {
    let redisClient;
    try {
        log('updateQueueRequestCount', '', 'update queue request count in redis');
        // initiate flag
        let isSimpleQueueRequest = true;
        // get redis client
        redisClient = getRedisInstance();

        // promisify redis functions
        const asyncGet = promisify(redisClient.get).bind(redisClient);
        const asyncWatch = promisify(redisClient.watch).bind(redisClient);
        const asyncMulti = (commands) => {
            const multi = redisClient.multi(commands);
            return promisify(multi.exec).call(multi);
        };

        // set retry count and delay
        let retryCount = 10;
        const retryDelay = 5000; // 5s
        const redisKey = `${studentAssignmentId}:${assignmentId}`;

        // redis watch
        const watchStarted = await asyncWatch(redisKey);
        log('updateQueueRequestCount', 'asyncWatch', watchStarted);
        while (retryCount > 0) {
            try {
                const currentVal = await asyncGet(redisKey);

                let newVal;
                if (!currentVal) {
                    newVal = {
                        requestCount: {
                            simple: isLiveClassEnabled ? 0 : 1,
                            priority: isLiveClassEnabled ? 1 : 0,
                        },
                    };
                    isSimpleQueueRequest = !isLiveClassEnabled;
                } else {
                    const parsedVal = JSON.parse(currentVal);
                    newVal = {
                        requestCount: {
                            simple: +parsedVal.requestCount.simple > 0 ? +parsedVal.requestCount.simple + 1 : 0,
                            priority: +parsedVal.requestCount.priority > 0 ? +parsedVal.requestCount.priority + 1 : 0,
                        },
                    };
                    isSimpleQueueRequest = +parsedVal.requestCount.simple > 0;
                }

                const updateRes = await asyncMulti(
                    [['set', redisKey, JSON.stringify(newVal)]],
                );
                log('updateQueueRequestCount', 'asyncMulti', updateRes);

                if (updateRes && updateRes[0] === 'OK') {
                    break;
                }
                if (updateRes === null) {
                    await sleep(retryDelay);
                }
            } catch (error) {
                console.error(error);
            }
            retryCount = retryCount - 1;
        }
        return isSimpleQueueRequest;
    } catch (error) {
        log('updateQueueRequestCount', 'ERROR', 'update queue request count in redis');
        console.error(error);
        throw error;
    } finally {
        log('updateQueueRequestCount', '', 'releasing redis connection');
        redisClient?.quit();
    }
};

const sleep = ms => new Promise(resolve => setTimeout(resolve, ms));

const pushMessageInQueue = async(isSimpleQueueRequest, message, studentAssignmentId, assignmentId) => {
    if (isSimpleQueueRequest) {
        log(
            new Date(),
            `pushMessageInQueue-start-${studentAssignmentId}:${assignmentId}`,
            `Message pushed in simple queue for ${studentAssignmentId}:${assignmentId}`,
        );
        // Push to simple queue
        if (process.env.STAGE === 'local') {
            const response: any = await dummyQueue(message, 'simple');
            log('dummyQueue', '', `Message Pushed In Simple Queue With Status Code ${ response.StatusCode || 500 }`);
        } else {
            const sqs = new SQS({ region: process.env.AWS_REGION });
            await sqs.sendMessage({
                MessageBody: JSON.stringify(message),
                QueueUrl: process.env.SAVE_STUDENT_ATTEMPTS_SIMPLE_SQS_URL,
                MessageGroupId: `${studentAssignmentId}`,
            }).promise();
        }
        log(
            new Date(),
            `pushMessageInQueue-stop-${studentAssignmentId}:${assignmentId}`,
            `Message pushed in simple queue for ${studentAssignmentId}:${assignmentId}`,
        );
    } else {
        log(
            new Date(),
            `pushMessageInQueue-start-${studentAssignmentId}:${assignmentId}`,
            `Message pushed in priority queue for ${studentAssignmentId}:${assignmentId}`,
        );
        // Push to priority queue
        if (process.env.STAGE === 'local') {
            const response: any = await dummyQueue(message, 'priority');
            log('dummyQueue', '', `Message Pushed In Priority Queue With Status Code ${ response.StatusCode || 500 }`);
        } else {
            const sqs = new SQS({ region: process.env.AWS_REGION });
            await sqs.sendMessage({
                MessageBody: JSON.stringify(message),
                QueueUrl: process.env.SAVE_STUDENT_ATTEMPTS_PRIORITY_SQS_URL,
                MessageGroupId: `${studentAssignmentId}`,
            }).promise();
        }
        log(
            new Date(),
            `pushMessageInQueue-stop-${studentAssignmentId}:${assignmentId}`,
            `Message pushed in priority queue for ${studentAssignmentId}:${assignmentId}`,
        );
    }
};

const saveSubmitResponse = async(data, requestId) => {
    const {
        isSubmitOrSave, activityId, studentId,
    } = data;
    const submitOrSave = isSubmitOrSave === SAVE ? 'SAVE' : 'SUBMIT';
    try {
        log(
            'saveSubmitResponse', '',
            `requestId-${requestId}: ${submitOrSave} response for user: ${studentId} and assignmentId: ${activityId}`,
        );
        const url = urlJoin(process.env.ASSIGNMENT_SUBMISSION_SERVICE, '/assignment-submission/save');
        const { data: { success, message = '' } } = await axios.post(
            url,
            { data },
            {
                httpsAgent: new https.Agent({ rejectUnauthorized: false }),
                headers: {
                    user_id: studentId,
                    request_id: requestId || uuidv4(),
                },
            },
        );
        return { success, message };
    } catch (error) {
        console.info(`${submitOrSave} action failed for user: ${studentId} and assignmentId: ${activityId}`);
        console.error(error);
        throw error;
    }
};

export default {
    checkLiveClassStatus,
    updateQueueRequestCount,
    pushMessageInQueue,
    saveSubmitResponse,
};
