import { camelizeKeys } from 'xcase';
import joi from '@hapi/joi';
import axios from 'axios';
import https from 'https';
import { v4 as uuidv4 } from 'uuid';
import urlJoin from 'url-join';
import { promisify } from 'util';
import { getRedisInstance } from '../../../redisManager';
import { Response as AssignmentResponse } from '../../../models/Response';
import { AssignmentTeacherStatus } from './../../../utils/AssignmentTeacherStatus';
import { AssignmentStudentStatus } from './../../../utils/AssignmentsStudentStatus';
import { StudentGradingStatus } from './../../../utils/StudentGradingStatus';
import { AssignmentStudentMapping } from './../../../models/AssignmentStudentMapping';
import { Assignments } from './../../../models/Assignments';
import { QuestionQuestionpartMap } from './../../../models/QuestionQuestionpartMap';
import { AssignmentsQuestionMap } from './../../../models/AssignmentsQuestionMap';
import { serverErrorCodes } from './../../../utils/ErrorCodes';
import { connection } from './../../../config/ormconfig';
import constants from './../../../utils/constants';

const {
    submitOrSave: { SAVE },
} = constants;

const studentPlayerViewSchema = (input) => {
    const gradingSchema = joi.object().keys({
        session_id: joi.string().uuid().allow(''),
        student_id: joi.string().uuid().required(),
        organisation_id: joi.string().uuid().required(),
        activity_id: joi.string().required(),
        activity_template_id: joi.string().required(),
        retrieve_tags: joi.boolean().allow(null),
    });
    return gradingSchema.validate(input);
};

const responseInputSchema = joi.object().keys({
    id: joi.number().optional().allow(null),
    xml_identifier: joi.string(),
    response_text: joi.string().allow(''),
});

const questionPartInputSchema = joi.object().keys({
    question_part_id: joi.number(),
    is_attempted: joi.boolean(),
    response: joi.array().items(responseInputSchema),
});

const questionInputSchema = joi.object().keys({
    question_id: joi.string().uuid(),
    question_parts: joi.array().items(questionPartInputSchema),
});

const saveSubmitInputSchema = (input) => {
    const saveSubmitSchema = joi.object().keys({
        session_id: joi.string().uuid(),
        student_id: joi.string().uuid().required(),
        organisation_id: joi.string().uuid().required(),
        activity_id: joi.string().required(),
        activity_template_id: joi.string().required(),
        questions: joi.array().items(questionInputSchema),
    });
    return saveSubmitSchema.validate(input);
};

const checkStudentExists = async (params) => {
    const {
        conn, sessionId, studentId, assessmentId, activityId: referenceId, organisationId,
    } = params;
    const query = await conn.manager.createQueryBuilder(AssignmentStudentMapping, 'student');
    query.select('student.id', 'id');
    query.leftJoin('student.assignments', 'assignments');
    query.where('assignments.referenceId = :referenceId', { referenceId });
    query.andWhere('assignments.assessmentId = :assessmentId', { assessmentId });
    query.andWhere('student.studentId = :studentId', { studentId });
    query.andWhere('student.sessionId = :sessionId', { sessionId });
    if (organisationId) {
        query.andWhere('assignments.organisationId = :organisationId', { organisationId });
    }
    return query.getRawOne();
};

const storeStudentData = async (params) => {
    const { assignmentId, studentId, sessionId, transaction } = params;
    const studentData = new AssignmentStudentMapping();
    studentData.assignmentId = assignmentId;
    studentData.gradingStatusId = StudentGradingStatus.NA;
    studentData.teacherStatusId = AssignmentTeacherStatus.ASSIGNED;
    studentData.studentStatusId = AssignmentStudentStatus.START;
    studentData.studentId = studentId;
    studentData.sessionId = sessionId;
    studentData.createdAt = new Date();
    return transaction.getRepository(AssignmentStudentMapping).save(studentData);
};

const getAssignmentDataByReferenceId = async (params) => {
    const { conn, assignmentReferenceId, organisationId, selectFields } = params;
    const assignment = conn.manager.createQueryBuilder(Assignments, 'assignment');
    assignment.where('assignment.referenceId = :assignmentReferenceId', { assignmentReferenceId });
    if (selectFields && selectFields?.length > 0) {
        assignment.select(selectFields);
    }
    if (organisationId) {
        assignment.andWhere('assignment.organisationId = :organisationId', { organisationId });
    }
    assignment.cache(true);
    return assignment.getOne();
};

const fetchAssignmentStudentResponse = async (params) => {
    const {
        conn, sessionId, studentId, assessmentId, activityId: referenceId, organisationId,
    } = params;
    const query = await conn.manager.createQueryBuilder(QuestionQuestionpartMap, 'questionPartResponse');
    query.select('questionPartResponse.question_part_id', 'questionPartId');
    query.addSelect('assignmentQuestionMap.question_id', 'questionId');
    query.addSelect('response.xmlIdentifier', 'xmlIdentifier');
    query.addSelect('response.responseText', 'responseText');
    query.addSelect('response.id', 'id');
    query.addSelect('questionPartResponse.is_attempted', 'isAttempted');
    query.leftJoin('questionPartResponse.response', 'response');
    query.leftJoin('questionPartResponse.assignmentQuestionMap', 'assignmentQuestionMap');
    query.leftJoin('assignmentQuestionMap.studentAssignment', 'studentAssignment');
    query.leftJoin('studentAssignment.assignments', 'assignments');
    query.where('assignments.referenceId = :referenceId', { referenceId });
    query.andWhere('assignments.assessmentId = :assessmentId', { assessmentId });
    query.andWhere('studentAssignment.studentId = :studentId', { studentId });
    query.andWhere('studentAssignment.sessionId = :sessionId', { sessionId });
    query.andWhere('assignments.organisation_id = :organisationId', { organisationId });
    return query.getRawMany();
};

const fetchAssignmentQuestionMap = async (params) => {
    const {
        conn, sessionId, studentId, assessmentId, activityId: referenceId, organisationId,
    } = params;
    const query = await conn.manager.createQueryBuilder(AssignmentsQuestionMap, 'assignmentQuestionMap');
    query.leftJoinAndSelect('assignmentQuestionMap.questionParts', 'questionParts');
    query.leftJoinAndSelect('questionParts.response', 'response');
    query.leftJoin('assignmentQuestionMap.studentAssignment', 'studentAssignment');
    query.leftJoin('studentAssignment.assignments', 'assignments');
    query.where('assignments.referenceId = :referenceId', { referenceId });
    query.andWhere('assignments.assessmentId = :assessmentId', { assessmentId });
    query.andWhere('studentAssignment.studentId = :studentId', { studentId });
    query.andWhere('studentAssignment.sessionId = :sessionId', { sessionId });
    query.andWhere('assignments.organisation_id = :organisationId', { organisationId });
    const data = await query.getMany();
    return camelizeKeys(data);
};

const extractQuestionPartsMap = (assignmentQuestionMaps) => {
    const questionQuestionPartmap = [];
    assignmentQuestionMaps.forEach(({ questionId, questionParts }) => {
        questionParts.forEach((qp) => {
            questionQuestionPartmap.push({
                ...qp,
                questionId,
            });
        });
    });
    return questionQuestionPartmap;
};

const storeAssignmentQuestionMaps = async (params) => {
    const {
        conn, studentAssignmentId, items, assignmentQuestionMaps, questionQuestionPartMaps,
        sessionId,
    } = params;
    let deleteQuestionIds = [];
    let insertedQuestionPartMaps = [];
    const insertAssignmentsQuestionMaps = [];
    for (const item of items) {
        const { questionId, questions } = item;
        const itemAssignmentQuestion = assignmentQuestionMaps.find(aqm => aqm.questionId === questionId);
        if (!itemAssignmentQuestion) {
            const questionPartMappings = [];
            questions.forEach((questionPart) => {
                const questionQuestionPartMap = new QuestionQuestionpartMap();
                questionQuestionPartMap.question_part_id = questionPart.id;
                questionQuestionPartMap.is_completely_graded = false;
                questionQuestionPartMap.is_attempted = false;
                questionQuestionPartMap.responseId = `${sessionId}_${uuidv4()}`;
                questionPartMappings.push(questionQuestionPartMap);
            });
            const assignmentQuestionMap = new AssignmentsQuestionMap();
            assignmentQuestionMap.question_id = questionId;
            assignmentQuestionMap.student_assignment_id = studentAssignmentId;
            assignmentQuestionMap.questionParts = questionPartMappings;
            insertAssignmentsQuestionMaps.push(assignmentQuestionMap);
        } else {
            const newQuestionPartMappings = [];
            const itemQuestionPartIds = questionQuestionPartMaps.filter(
                qpm => qpm.assignmentQuestionMappingId === itemAssignmentQuestion.id,
            ).map(q => q.questionPartId);
            const questionIds = questions.map(q => q.id);
            const newQuestions = questions.filter(q => !itemQuestionPartIds.includes(q.id));

            const deletedQuestions = questionQuestionPartMaps.filter(qpm => (
                qpm.assignmentQuestionMappingId === itemAssignmentQuestion.id
                && !questionIds.includes(qpm.questionPartId)
            )).map(q => q.id);
            deleteQuestionIds = [...deleteQuestionIds, ...deletedQuestions];

            newQuestions.forEach((questionPart) => {
                const questionQuestionPartMap = new QuestionQuestionpartMap();
                questionQuestionPartMap.question_part_id = questionPart.id;
                questionQuestionPartMap.is_completely_graded = false;
                questionQuestionPartMap.is_attempted = false;
                questionQuestionPartMap.responseId = `${sessionId}_${uuidv4()}`;
                newQuestionPartMappings.push(questionQuestionPartMap);
            });

            if (newQuestionPartMappings.length) {
                itemAssignmentQuestion.questionParts =
                    [...itemAssignmentQuestion.questionParts, ...newQuestionPartMappings];
                insertAssignmentsQuestionMaps.push(itemAssignmentQuestion);
            }
        }
    }
    if (insertAssignmentsQuestionMaps.length) {
        await conn.manager.getRepository(AssignmentsQuestionMap).save(insertAssignmentsQuestionMaps);
        insertedQuestionPartMaps = extractQuestionPartsMap(camelizeKeys(insertAssignmentsQuestionMaps));
    }

    if (deleteQuestionIds.length) {
        await conn.createQueryBuilder().delete().from(QuestionQuestionpartMap)
            .where('id IN (:...deleteQuestionIds)', { deleteQuestionIds })
            .execute();
    }
    return insertedQuestionPartMaps;
};

const fetchAssessmentCorrectAnswerView = async (params) => {
    const { organisationId, activityTemplateId: activityReference } = params;
    const data = { activityReference, organisationId };
    const url = urlJoin(process.env.ACTIVITY_SERVICE, '/activity/view/correctAnswer');
    const response = await axios.post(
        url,
        data,
        {
            httpsAgent: new https.Agent({ rejectUnauthorized: false }),
        });
    return response.data.activityData;
};

const doesStudentHaveAccessToAssignment = async (params) => {
    const { assignmentId, studentId } = params;
    const con = await connection;
    if (!assignmentId || !studentId) {
        return null;
    }
    try {
        const query = con.manager
        .createQueryBuilder(Assignments, 'assignment')
        .innerJoin(
            AssignmentStudentMapping,
            'studentAssignment', 'studentAssignment.assignmentId = assignment.id')
        .where('assignment.assignmentId = :assignmentId', { assignmentId })
        .andWhere('student_id = :studentId', { studentId })
        .select([
            'studentAssignment.id',
            'assignment.assessmentId',
        ])
        .cache(true);
        const rawData = await query.getRawMany();

        if (rawData && rawData.length === 1) {
            return camelizeKeys(rawData[0]);
        }
        return null;
    } catch (error) {
        console.error(error);
        throw Object.assign(
            new Error(error),
            { ...serverErrorCodes.DATABASE_ERROR, originalMessage: error.message, stack: error.stack },
        );
    }
};

const doesStudentHasAccessOrIsAuthorized = async(params) => {
    const studentHasAccess = await doesStudentHaveAccessToAssignment(params);
    if (!studentHasAccess) {
        throw serverErrorCodes.NO_ACCESS;
    }
    return studentHasAccess;
};

const getResponseArray = (params) => {
    const { responseTexts } = params;

    let newResponses;
    return responseTexts.map((response) => {
        const { xmlIdentifier, responseText } = response;
        if (responseText !== '' && !responseText) {
            return null;
        }
        newResponses = new AssignmentResponse();
        newResponses.xmlIdentifier = xmlIdentifier;
        newResponses.responseText = responseText;
        return newResponses;
    });
};

const updateOrInsertQuestionData = async(con, queryData) => {
    const NEW_QUESTION = 1;
    const NEW_QUESTION_PART = 2;
    let result = null;
    const {
        newItemToSubmit, questionPartId, responseArray, questionId, foreignKey,
        isAttempted, sessionId,
    } = queryData;

    if (newItemToSubmit === NEW_QUESTION) {
        const questionQuestionPartMap = new QuestionQuestionpartMap();
        questionQuestionPartMap.question_part_id = questionPartId;
        questionQuestionPartMap.is_completely_graded = false;
        questionQuestionPartMap.response = responseArray;
        questionQuestionPartMap.is_attempted = isAttempted;
        questionQuestionPartMap.responseId = `${sessionId}_${uuidv4()}`;
        const assignmentQuestionMap = new AssignmentsQuestionMap();
        assignmentQuestionMap.question_id = questionId;
        assignmentQuestionMap.student_assignment_id = foreignKey;
        assignmentQuestionMap.questionParts = [questionQuestionPartMap];

        result = await con.manager.getRepository(AssignmentsQuestionMap).save(assignmentQuestionMap);
        result = result.questionParts[0].id;
    }

    if (newItemToSubmit === NEW_QUESTION_PART) {
        const questionQuestionPartMap = new QuestionQuestionpartMap();
        questionQuestionPartMap.question_part_id = questionPartId;
        questionQuestionPartMap.assignment_question_mapping_id = foreignKey;
        questionQuestionPartMap.is_completely_graded = false;
        questionQuestionPartMap.response = responseArray;
        questionQuestionPartMap.is_attempted = isAttempted;
        questionQuestionPartMap.responseId = `${sessionId}_${uuidv4()}`;

        result = await con.manager.getRepository(QuestionQuestionpartMap).save(questionQuestionPartMap);
        result = result.id;
    }
    return result;
};

const insertResponses = async (insertData) => {
    const con = await connection;
    const {
        responseTextsArray: responseTexts, response: item, foreignKey, newItemToSubmit,
        isSubmitOrSave, isEmptyMultipleChoiceResponse, isAttempted, sessionId,
    } = insertData;
    if (!item || !foreignKey) {
        return null;
    }

    const { questionId, questionPartId } = item;

    const responseArray = getResponseArray({ responseTexts });
    if ((responseArray.every(response => !response) || isEmptyMultipleChoiceResponse) &&
        isSubmitOrSave === SAVE) {
        return;
    }

    try {
        let result = null;
        const queryData = {
            newItemToSubmit, questionPartId, responseArray, questionId, foreignKey,
            isAttempted, sessionId,
        };
        result = await updateOrInsertQuestionData(con, queryData);
        return result;
    } catch (error) {
        if (parseInt(error.code, 10) === 23505) {
            console.error(error);
            return null;
        }
        console.error(error);
        throw Object.assign(
            new Error(error),
            { ...serverErrorCodes.DATABASE_ERROR, originalMessage: error.message, stack: error.stack },
        );
    }
};

const getPointsScoredAndIsGraded = (score) => {
    if (Number(score) === 0 && score !== null) {
        return { isCompletelyGraded: true, pointsScored: 0 };
    }
    const pointsScored = Number(score) || null;
    if (pointsScored === null) {
        return { pointsScored, isCompletelyGraded: false };
    }
    return { pointsScored, isCompletelyGraded: true };
};

const getActivityData = async(assessmentParamas) => {
    let redisClient;
    let activityData;
    try {
        const { activityTemplateId, organisationId, timeLogger } = assessmentParamas;

        timeLogger.start('get-redis-connection');
        redisClient = getRedisInstance();
        timeLogger.stop('get-redis-connection');

        const asyncGet = promisify(redisClient.get).bind(redisClient);
        const redisKey = `fetchAssessmentCorrectAnswerView:${activityTemplateId}:${organisationId}`;

        timeLogger.start('read-fetchAssessmentCorrectAnswerView-from-cache');
        activityData = await asyncGet(redisKey);
        timeLogger.stop('read-fetchAssessmentCorrectAnswerView-from-cache');

        if (activityData) {
            activityData = JSON.parse(activityData || null);
        }

        if (!activityData) {
            timeLogger.start('read-fetchAssessmentCorrectAnswerView-from-db');
            activityData = await fetchAssessmentCorrectAnswerView(assessmentParamas);
            timeLogger.stop('read-fetchAssessmentCorrectAnswerView-from-db');

            const asyncSet = promisify(redisClient.set).bind(redisClient);
            timeLogger.start('write-fetchAssessmentCorrectAnswerView-into-cache');
            asyncSet(redisKey, JSON.stringify(activityData), 'EX', process.env.PLAYER__PLAYER__TYPEORM_CACHE_TYPE);
            timeLogger.stop('write-fetchAssessmentCorrectAnswerView-into-cache');
        }

        return activityData;
    } catch (error) {
        console.error(error);
        return activityData;
    } finally {
        redisClient?.quit();
    }
};

export default {
    studentPlayerViewSchema,
    fetchAssignmentStudentResponse,
    fetchAssessmentCorrectAnswerView,
    saveSubmitInputSchema,
    checkStudentExists,
    storeStudentData,
    getAssignmentDataByReferenceId,
    doesStudentHasAccessOrIsAuthorized,
    insertResponses,
    getResponseArray,
    getPointsScoredAndIsGraded,
    fetchAssignmentQuestionMap,
    storeAssignmentQuestionMaps,
    extractQuestionPartsMap,
    getActivityData,
    doesStudentHaveAccessToAssignment,
};
